use std::{
    error::Error,
    path::{Path, PathBuf},
};

use log::{debug, info};

use radicle::prelude::Profile;
use radicle_ci_broker::{
    adapter::Adapter,
    broker::Broker,
    config::Config,
    error::BrokerError,
    event::{BrokerEvent, NodeEventSource},
    msg::Request,
    run::Run,
    status::StatusBuilder,
};

fn main() {
    if let Err(e) = fallible_main() {
        eprintln!("ERROR: {}", e);
        let mut e = e.source();
        while let Some(source) = e {
            eprintln!("caused by: {}", source);
            e = source.source();
        }
    }
}

fn fallible_main() -> Result<(), BrokerError> {
    pretty_env_logger::init();
    info!("Radicle CI broker starts");

    let mut args = std::env::args().skip(1);
    let filename: PathBuf = if let Some(filename) = args.next() {
        PathBuf::from(filename)
    } else {
        return Err(BrokerError::Usage);
    };

    let config = Config::load(&filename)?;
    debug!("loaded configuration: {:#?}", config);

    let mut broker = Broker::default();
    debug!("created broker");

    // FIXME: this is broken. the config file only lists how to invoke
    // each adapter, not what adapter to use for each repo.
    // for (rid, spec) in config.adapters.iter() {
    //     debug!("setting adapter for {rid:?} to {spec:#?}");
    //     let rid = RepoId::from_urn(rid).map_err(|e| BrokerError::BadRepoId(rid.into(), e))?;
    //     let adapter = Adapter::new(&spec.command).with_environment(spec.envs());
    //     broker.set_repository_adapter(&rid, &adapter);
    // }
    // debug!("set per-repository adapters");

    let spec =
        config
            .adapter(&config.default_adapter)
            .ok_or(BrokerError::UnknownDefaultAdapter(
                config.default_adapter.clone(),
            ))?;
    let adapter = Adapter::new(&spec.command).with_environment(spec.envs());
    broker.set_default_adapter(&adapter);
    debug!("set default adapter");

    let profile = Profile::load()?;
    debug!("loaded profile");

    let mut source = NodeEventSource::new(&profile)?;
    debug!("created node event source");

    for filter in config.filters.iter() {
        source.allow(filter.clone());
    }
    debug!("added filters to node event source");

    // This loop ends when there's an error, e.g., failure to read an
    // event from the node.
    let mut counter = 0;
    let mut status = StatusBuilder::new(config.status_page().unwrap_or(Path::new("/dev/null")));
    loop {
        status.write()?;
        debug!("waiting for event from node");
        for e in source.event()? {
            status.broker_event(&e);
            status.write()?;

            counter += 1;
            debug!("broker event {e:#?}");
            let BrokerEvent::RefChanged {
                rid,
                oid,
                name: _,
                old: _,
            } = e;

            let req = Request::trigger(&profile, &e)?;
            let mut run = Run::default();
            if let Some(adapter) = broker.adapter(&rid) {
                adapter.run(&req, &mut run)?;
                status.ci_run(run.adapter_run_id().unwrap(), run.result().unwrap());
                println!(
                    "Run CI run #{}: {}, {} -> {}",
                    counter,
                    rid,
                    oid,
                    run.result().unwrap()
                );
            } else {
                eprintln!("ERROR: no adapter available for repository {rid}: not running CI");
            }
        }
    }
}
