use std::path::{Path, PathBuf};

use serde::Serialize;
use time::{macros::format_description, OffsetDateTime};

use crate::{
    event::BrokerEvent,
    msg::{RunId, RunResult},
};

#[derive(Serialize)]
pub struct Status {
    timestamp: String,
    ci_broker_version: &'static str,
    ci_broker_git_commit: &'static str,
    latest_broker_event: Option<BrokerEvent>,
    latest_ci_run: Option<RunId>,
    latest_ci_run_result: Option<RunResult>,
}

impl Status {
    pub fn write(&self, filename: &Path) -> Result<(), StatusError> {
        let s = serde_json::to_string_pretty(&self).map_err(StatusError::serialize)?;
        std::fs::write(filename, s.as_bytes())
            .map_err(|e| StatusError::status_write(filename, e))?;
        Ok(())
    }
}

pub struct StatusBuilder {
    filename: PathBuf,
    latest_broker_event: Option<BrokerEvent>,
    latest_ci_run: Option<RunId>,
    latest_ci_run_result: Option<RunResult>,
}

impl StatusBuilder {
    pub fn new(filename: &Path) -> Self {
        Self {
            filename: filename.into(),
            latest_ci_run: None,
            latest_ci_run_result: None,
            latest_broker_event: None,
        }
    }

    pub fn broker_event(&mut self, event: &BrokerEvent) {
        self.latest_broker_event = Some(event.clone());
    }

    pub fn ci_run(&mut self, run_id: &RunId, result: &RunResult) {
        self.latest_ci_run = Some(run_id.clone());
        self.latest_ci_run_result = Some(result.clone());
    }

    pub fn write(&self) -> Result<(), StatusError> {
        let status = Status {
            timestamp: Self::now()?,
            ci_broker_version: env!("CARGO_PKG_VERSION"),
            ci_broker_git_commit: env!("GIT_HEAD"),
            latest_broker_event: self.latest_broker_event.clone(),
            latest_ci_run: self.latest_ci_run.clone(),
            latest_ci_run_result: self.latest_ci_run_result.clone(),
        };
        status.write(&self.filename)?;
        Ok(())
    }

    fn now() -> Result<String, StatusError> {
        let fmt = format_description!("[year]-[month]-[day] [hour]:[minute]:[second]Z");
        OffsetDateTime::now_utc()
            .format(fmt)
            .map_err(StatusError::format_now)
    }
}

#[derive(Debug, thiserror::Error)]
pub enum StatusError {
    #[error("failed to format current time stamp")]
    FormatNow(#[source] time::error::Format),

    #[error("failed to serialize status as JSON")]
    Serizalize(#[source] serde_json::Error),

    #[error("failed to write status to file {0}")]
    StatusWrite(PathBuf, #[source] std::io::Error),
}

impl StatusError {
    fn format_now(err: time::error::Format) -> Self {
        Self::FormatNow(err)
    }

    fn serialize(err: serde_json::Error) -> Self {
        Self::Serizalize(err)
    }

    fn status_write(filename: &Path, err: std::io::Error) -> Self {
        Self::StatusWrite(filename.into(), err)
    }
}
